import ansys.fluent.core.quantity as q


class Quantity(float):
    """Quantity instantiates physical quantities using their real values and
    units. All the instances of this class are converted to base SI units system to have
    consistency in arithmetic operations.

    Parameters
    ----------
    value : int | float
        Real value of quantity.
    units : str
        Unit string representation of quantity.
    quantity_map : dict
        Quantity map representation of quantity.
    dimensions : list
        Dimensions representation of quantity.

    Methods
    -------
    to()
        Converts to given unit string.

    Returns
    -------
    Quantity instance.
    """

    def __new__(cls, value, units=None, quantity_map=None, dimensions=None):
        if (
            (units and quantity_map)
            or (units and dimensions)
            or (quantity_map and dimensions)
        ):
            raise QuantityError.EXCESSIVE_PARAMETERS()

        _units_table = q.UnitsTable()
        _value = float(value)

        if units is not None:
            _unit = units

        if quantity_map:
            units = q.QuantityMap(quantity_map).units
            _unit = units

        if dimensions:
            _dimensions = q.Dimensions(dimensions=dimensions)
            _unit = _dimensions.units

        _, si_multiplier, si_offset = _units_table.si_data(units=_unit)
        _si_value = (_value + si_offset) * si_multiplier

        return float.__new__(cls, _si_value)

    def __init__(self, value, units=None, quantity_map=None, dimensions=None):
        if (
            (units and quantity_map)
            or (units and dimensions)
            or (quantity_map and dimensions)
        ):
            raise QuantityError.EXCESSIVE_PARAMETERS()

        self._units_table = q.UnitsTable()
        self._value = float(value)

        if units is not None:
            self._unit = units
            self._dimensions = q.Dimensions(units=units)

        if quantity_map:
            units = q.QuantityMap(quantity_map).units
            self._unit = units
            self._dimensions = q.Dimensions(units=units)

        if dimensions:
            self._dimensions = q.Dimensions(dimensions=dimensions)
            self._unit = self._dimensions.units

        self._type = self._units_table.get_type(self._unit)

        si_units, si_multiplier, si_offset = self._units_table.si_data(units=self._unit)

        self._si_units = si_units[:-1]
        self._si_value = (self.value + si_offset) * si_multiplier

    def _arithmetic_precheck(self, __value, caller=None) -> str:
        """Validate dimensions of quantities.

        Parameters
        ----------
        __value : Quantity | int | float
            Value modifying current quantity object.
        caller : str
            Name of caller function.
        Returns
        -------
        str
            SI unit string of new quantity.
        """
        # Cannot perform operations between quantities with incompatible dimensions
        if isinstance(__value, Quantity) and (self.dimensions != __value.dimensions):
            raise QuantityError.INCOMPATIBLE_DIMENSIONS(self.units, __value.units)
        # Cannot perform operations on a non-dimensionless quantity
        if (
            caller not in ["__mul__", "__truediv__"]
            and (not self.is_dimensionless)
            and (not isinstance(__value, Quantity))
            and isinstance(__value, (float, int))
        ):
            raise QuantityError.INCOMPATIBLE_VALUE(__value)

        return (
            "delta_K"
            if self.type in ["Temperature", "Temperature Difference"]
            else self.si_units
        )

    @property
    def value(self):
        """Real value."""
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value

    @property
    def units(self):
        """Unit string."""
        return self._unit

    @property
    def si_value(self):
        """SI conversion value."""
        return self._si_value

    @property
    def si_units(self):
        """SI conversion unit string."""
        return self._si_units

    @property
    def dimensions(self):
        """Dimensions."""
        return self._dimensions.dimensions

    @property
    def is_dimensionless(self) -> bool:
        """Check if quantity is dimensionless."""
        return all([dim == 0.0 for dim in self.dimensions])

    @property
    def type(self):
        """Type."""
        return self._type

    def to(self, to_units: str) -> "Quantity":
        """Perform quantity conversions.

        Parameters
        ----------
        to_units : str
            Desired unit to be converted to.

        Returns
        -------
        Quantity
            Quantity object containing desired quantity conversion.
        """

        if not isinstance(to_units, str):
            raise TypeError("`to_units` should be a `str` type.")

        # Retrieve all SI required SI data and perform conversion
        _, si_multiplier, si_offset = self._units_table.si_data(to_units)
        new_value = (self.si_value / si_multiplier) - si_offset

        new_obj = Quantity(value=new_value, units=to_units)

        # Confirm conversion compatibility
        self._arithmetic_precheck(new_obj)

        return new_obj

    def __str__(self):
        return f'({self.value}, "{self.units}")'

    def __repr__(self):
        return f'Quantity ({self.value}, "{self.units}")'

    def __pow__(self, __value):
        temp_dimensions = [dim * __value for dim in self.dimensions]
        new_si_value = self.si_value**__value
        new_dimensions = q.Dimensions(dimensions=temp_dimensions)
        return Quantity(value=new_si_value, units=new_dimensions.units)

    def __mul__(self, __value):
        new_units = self._arithmetic_precheck(__value, caller="__mul__")
        if isinstance(__value, Quantity):
            temp_dimensions = [
                dim + __value.dimensions[idx] for idx, dim in enumerate(self.dimensions)
            ]
            new_si_value = self.si_value * __value.si_value
            new_dimensions = q.Dimensions(dimensions=temp_dimensions)
            new_units = new_units if new_units == "delta_K" else new_dimensions.units
            return Quantity(value=new_si_value, units=new_units)

        if isinstance(__value, (float, int)):
            return Quantity(value=self.si_value * __value, units=new_units)

    def __rmul__(self, __value):
        return self.__mul__(__value)

    def __truediv__(self, __value):
        new_units = self._arithmetic_precheck(__value, caller="__truediv__")
        if isinstance(__value, Quantity):
            temp_dimensions = [
                dim - __value.dimensions[idx] for idx, dim in enumerate(self.dimensions)
            ]
            new_si_value = self.si_value / __value.si_value
            new_dimensions = q.Dimensions(dimensions=temp_dimensions)
            new_units = new_units if new_units == "delta_K" else new_dimensions.units
            return Quantity(value=new_si_value, units=new_units)

        if isinstance(__value, (float, int)):
            return Quantity(value=self.si_value / __value, units=new_units)

    def __rtruediv__(self, __value):
        return self.__truediv__(__value)

    def __add__(self, __value):
        new_units = self._arithmetic_precheck(__value)
        new_value = float(self) + float(__value)
        return Quantity(value=new_value, units=new_units)

    def __radd__(self, __value):
        return self.__add__(__value)

    def __sub__(self, __value):
        new_units = self._arithmetic_precheck(__value)
        new_value = float(self) - float(__value)
        return Quantity(value=new_value, units=new_units)

    def __rsub__(self, __value):
        return self.__sub__(__value)

    def __neg__(self):
        return Quantity(-self.value, self.units)

    def __gt__(self, __value):
        self._arithmetic_precheck(__value)
        return float(self) > float(__value)

    def __ge__(self, __value):
        self._arithmetic_precheck(__value)
        return float(self) >= float(__value)

    def __lt__(self, __value):
        self._arithmetic_precheck(__value)
        return float(self) < float(__value)

    def __le__(self, __value):
        self._arithmetic_precheck(__value)
        return float(self) <= float(__value)

    def __eq__(self, __value):
        self._arithmetic_precheck(__value)
        return float(self) == float(__value)

    def __neq__(self, __value):
        self._arithmetic_precheck(__value)
        return float(self) != float(__value)


class QuantityError(ValueError):
    """Custom quantity errors."""

    def __init__(self, err):
        super().__init__(err)

    @classmethod
    def EXCESSIVE_PARAMETERS(cls):
        return cls(
            "Quantity only accepts 1 of the following parameters: (units) or (quantity_map) or (dimensions)."
        )

    @classmethod
    def INCOMPATIBLE_DIMENSIONS(cls, from_unit, to_unit):
        return cls(f"`{from_unit}` and `{to_unit}` have incompatible dimensions.")

    @classmethod
    def INCOMPATIBLE_VALUE(cls, value):
        return cls(f"`{value}` is incompatible with the current quantity object.")



import ansys.fluent.core.quantity as q


class UnitSystem:
    """Initializes a unit system based on user-defined units or a pre-definned unit system.

    Parameters
    ----------
    name: str
        Custom name associated with a user-defined unit system.
    base_units: list
        Custom units associated with a user-defined unit system.
    unit_sys: str
        Pre-defined unit system.

    Methods
    -------
    convert()
        Converts from one unit system to to given unit system.

    Returns
    -------
    Quantity instance.
    """

    def __init__(self, name: str = None, base_units: list = None, unit_sys: str = None):
        self._units_table = q.UnitsTable()

        if name and unit_sys or base_units and unit_sys:
            raise UnitSystemError.EXCESSIVE_PARAMETERS()

        if base_units:
            if len(base_units) != q.Dimensions.max_dim_len():
                raise UnitSystemError.BASE_UNITS_LENGTH(len(base_units))

            for idx, unit in enumerate(base_units):
                if unit not in self._units_table.fundamental_units:
                    raise UnitSystemError.UNIT_UNDEFINED(unit)

                if (idx + 1) != self._units_table.dimension_order[
                    self._units_table.fundamental_units[unit]["type"]
                ]:
                    raise UnitSystemError.UNIT_ORDER(
                        t1=list(self._units_table.dimension_order.keys())[idx],
                        o1=idx + 1,
                        t2=self._units_table.fundamental_units[unit]["type"],
                        o2=self._units_table.dimension_order[
                            self._units_table.fundamental_units[unit]["type"]
                        ],
                    )

            self._name = name
            self._base_units = base_units

        if unit_sys is not None:
            if unit_sys not in self._units_table.unit_systems:
                raise UnitSystemError.INVALID_UNIT_SYS(unit_sys)

            self._name = unit_sys
            self._base_units = self._units_table.unit_systems[unit_sys]

    def convert(self, quantity: q.Quantity) -> q.Quantity:
        """Perform unit system conversions.

        Parameters
        ----------
        quantity : Quantity
            Desired quantity object to convert.

        Returns
        -------
        Quantity
            Quantity object containing desired unit system conversion.
        """

        new_dim = q.Dimensions(
            dimensions=quantity.dimensions, unit_sys=self._base_units
        )

        return q.Quantity(value=quantity.value, units=new_dim.units)

    @property
    def name(self):
        """Name associated with unit system."""
        return self._name

    @property
    def base_units(self):
        """Units associated with unit system."""
        return self._base_units


class UnitSystemError(ValueError):
    """Custom unit system errors."""

    def __init__(self, err):
        super().__init__(err)

    @classmethod
    def EXCESSIVE_PARAMETERS(cls):
        return cls(
            "UnitSystem only accepts 1 of the following parameters: (name, base_units) or (unit_sys)."
        )

    @classmethod
    def BASE_UNITS_LENGTH(cls, len):
        return cls(
            f"The `base_units` argument must contain 9 units, currently there are {len}."
        )

    @classmethod
    def UNIT_UNDEFINED(cls, unit):
        return cls(
            f"`{unit}` is an undefined unit. To use `{unit}` add it to the `fundamental_units` table within quantity_config.yaml."
        )

    @classmethod
    def UNIT_ORDER(cls, t1, o1, t2, o2):
        return cls(
            f"Expected unit of type: `{t1}` (order: {o1}), received unit of type: `{t2}` (order: {o2})."
        )

    @classmethod
    def INVALID_UNIT_SYS(cls, sys):
        return cls(f"`{sys}` is not a supported unit system.")



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _HasAllowedValuesMixin

class surface_tension_scale(Boolean):
    """
    Include surface tension based time scale.
    """

    fluent_name = "surface-tension-scale?"




import ansys.fluent.core.quantity as q


class QuantityMap(object):
    """Creates a Quantity Map object based on a given quantity map.

    Parameters
    ----------
    quantity_map : dict
        Dictionary containing quantity map units and values.

    Returns
    -------
    QuantityMap instance.
    """

    def __init__(self, quantity_map):
        self._units_table = q.UnitsTable()

        for item in quantity_map:
            if item not in self._units_table.api_quantity_map:
                raise QuantityMapError.UNKNOWN_MAP_ITEM(item)

        self._units = self._map_to_units(quantity_map)

    def _map_to_units(self, quantity_map: dict) -> str:
        """Convert a quantity map into a unit string.

        Parameters
        ----------
        quantity_map : dict
            Quantity map to be converted to unit string.

        Returns
        -------
        str
            Unit string representation of quantity map.
        """
        unit_dict = {
            self._units_table.api_quantity_map[term]: power
            for term, power in quantity_map.items()
        }

        units = ""

        # Split unit string into terms and parse data associated with individual terms
        for terms in unit_dict:
            for term in terms.split(" "):
                _, unit_term, unit_term_power = self._units_table.filter_unit_term(term)

                unit_term_power *= unit_dict[terms]

                if unit_term_power == 1.0:
                    units += f"{unit_term} "
                elif unit_term_power != 0.0:
                    units += f"{unit_term}^{unit_term_power} "

        return self._units_table.condense(units=units)[:-1]

    @property
    def units(self):
        """Unit string representation of quantity map."""
        return self._units


class QuantityMapError(ValueError):
    """Custom quantity map errors."""

    def __init__(self, err):
        super().__init__(err)

    @classmethod
    def UNKNOWN_MAP_ITEM(cls, item):
        return cls(f"`{item}` is not a valid quantity map item.")



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _HasAllowedValuesMixin

class barycentric_sampling_enabled(Boolean):
    """
    Enable/disable using of barycentric coordinates when sampling particles at planes.
    """

    fluent_name = "barycentric-sampling-enabled?"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _HasAllowedValuesMixin

class calculator_method(String, _HasAllowedValuesMixin):
    """
    Set a method for solar calculator.
    """

    fluent_name = "calculator-method"




#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _HasAllowedValuesMixin

from .option_2 import option as option_cls
from .zero_shear_viscosity import zero_shear_viscosity as zero_shear_viscosity_cls
from .power_law_index import power_law_index as power_law_index_cls
from .time_constant import time_constant as time_constant_cls
from .reference_temperature import reference_temperature as reference_temperature_cls
from .activation_energy import activation_energy as activation_energy_cls
class cross(Group):
    """
    'cross' child.
    """

    fluent_name = "cross"

    child_names = \
        ['option', 'zero_shear_viscosity', 'power_law_index', 'time_constant',
         'reference_temperature', 'activation_energy']

    option: option_cls = option_cls
    """
    option child of cross.
    """
    zero_shear_viscosity: zero_shear_viscosity_cls = zero_shear_viscosity_cls
    """
    zero_shear_viscosity child of cross.
    """
    power_law_index: power_law_index_cls = power_law_index_cls
    """
    power_law_index child of cross.
    """
    time_constant: time_constant_cls = time_constant_cls
    """
    time_constant child of cross.
    """
    reference_temperature: reference_temperature_cls = reference_temperature_cls
    """
    reference_temperature child of cross.
    """
    activation_energy: activation_energy_cls = activation_energy_cls
    """
    activation_energy child of cross.
    """



#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import _ChildNamedObjectAccessorMixin

from ansys.fluent.core.solver.flobject import _CreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _NonCreatableNamedObjectMixin

from ansys.fluent.core.solver.flobject import _HasAllowedValuesMixin

class list_surfaces_by_color(Command):
    """
    To list the surfaces by its color.
    """

    fluent_name = "list-surfaces-by-color"

